/*
 * $Id: AbstractTTreeNode.java,v 1.1 2005/06/30 01:14:44 ahimanikya Exp $
 * ======================================================================= Copyright (c)
 * 2002-2004 Axion Development Team. All rights reserved. Redistribution and use in source
 * and binary forms, with or without modification, are permitted provided that the
 * following conditions are met: 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other materials
 * provided with the distribution. 3. The names "Tigris", "Axion", nor the names of its
 * contributors may not be used to endorse or promote products derived from this software
 * without specific prior written permission. 4. Products derived from this software may
 * not be called "Axion", nor may "Tigris" or "Axion" appear in their names without
 * specific prior written permission. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.ext.indexes.ttree;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import org.axiondb.AxionException;
import org.axiondb.io.AxionFileSystem;

/**
 * Base Node implementation for T-Tree.
 * 
 * @version $Revision: 1.1 $ $Date: 2005/06/30 01:14:44 $
 * @author Pavels Andrejevs
 */

public abstract class AbstractTTreeNode {

    protected static AxionFileSystem FS = new AxionFileSystem();

    /**
     * Return a String comprised of N spaces.
     */
    protected static final String space(int n) {
        StringBuffer buf = new StringBuffer(0);
        for (int i = 0; i < n; i++) {
            buf.append("  ");
        }
        return buf.toString();
    }
    
    protected int _balance;
    protected int _fileId;
    protected AbstractTTreeNode _left;
    protected AbstractTTreeNode _next;
    protected int _nodeSize;
    protected AbstractTTreeNode _prev;
    protected AbstractTTreeNode _right;
    protected TTree _tree;
    private boolean _dirty;

    public AbstractTTreeNode(TTree tree, int fileId) throws AxionException {
        _tree = tree;
        if (tree.isMemoryOnly()) {
            _dirty = true;
        }
        if (fileId != 0) {
            // Load existing node from disk.
            setFileId(fileId);
            read();
        } else {
            // A new node is just created.
            setFileId(tree.getMetaData().incrementCounter());
            create();
        }
    }

    public abstract void addKeyValue(Object key, int value) throws AxionException;

    public void clearDirty() {
        _dirty = false;
    }

    public abstract void copyKeysValues(int from, int to, int length) throws AxionException;

    // Return value:
    // 1: The tree diminished in height.
    // 0: The tree didn't change in height.
    // -1: The key was not found.
    public final int delete(Object key, int value, NodeReference ref) throws AxionException {
        AbstractTTreeNode pg;
        int n = _nodeSize;
        int diff = _tree.compare(key, getKey(0));
        if (diff <= 0) {
            if (_left != null) {
                pg = ref.pg;
                ref.pg = _left;
                int h = _left.delete(key, value, ref);
                if (_left != ref.pg) {
                    _left = ref.pg;
                    markDirty();
                }
                ref.pg = pg;
                if (h > 0) {
                    return balanceLeftBranch(ref, false);
                } else if (h == 0) {
                    return 0;
                }
            }
        }
        diff = _tree.compare(key, getKey(n - 1));
        if (diff <= 0) {
            for (int i = 0; i < n; i++) {
                if (getValue(i) == value) {
                    if (n == 1) {
                        if (_right == null || _left == null) {
                            // delete this
                            if (_right == null) {
                                ref.pg = _left;
                            } else {
                                ref.pg = _right;
                            }
                            unlinkThis();
                            _tree.deleteNode(this);
                            return 1;
                        }
                    }
                    if (n <= _tree.getMinNodeSize()) {
                        if (_left != null && _balance <= 0) {
                            AbstractTTreeNode prev = getPrev();
                            if (i > 0) {
                                copyKeysValues(0, 1, i);
                            }

                            Object tKey = prev.getKey(prev._nodeSize - 1);
                            int tValue = prev.getValue(prev._nodeSize - 1);
                            setKeyValue(0, tKey, tValue);
                            pg = ref.pg;
                            ref.pg = _left;
                            int h = _left.delete(tKey, tValue, ref);
                            if (_left != ref.pg) {
                                _left = ref.pg;
                                markDirty();
                            }
                            ref.pg = pg;
                            if (h > 0) {
                                h = balanceLeftBranch(ref, false);
                            }
                            return h;
                        } else if (_right != null) {
                            AbstractTTreeNode next = getNext();
                            int len = n - i - 1;
                            if (len > 0) {
                                copyKeysValues(i + 1, i, len);
                            }
                            Object tKey = next.getKey(0);
                            int tValue = next.getValue(0);
                            setKeyValue(n - 1, tKey, tValue);
                            pg = ref.pg;
                            ref.pg = _right;
                            int h = _right.delete(tKey, tValue, ref);
                            if (_right != ref.pg) {
                                _right = ref.pg;
                                markDirty();
                            }
                            ref.pg = pg;
                            if (h > 0) {
                                h = balanceRightBranch(ref, false);
                            }
                            return h;
                        }
                    }
                    int len = n - i - 1;
                    if (len > 0) {
                        copyKeysValues(i + 1, i, len);
                    }
                    removeKeyValue(n - 1);
                    return 0;
                }
            }
        }
        if (_right != null) {
            pg = ref.pg;
            ref.pg = _right;
            int h = _right.delete(key, value, ref);
            if (_right != ref.pg) {
                _right = ref.pg;
                markDirty();
            }
            ref.pg = pg;
            if (h > 0) {
                return balanceRightBranch(ref, false);
            } else {
                return h;
            }
        }
        return -1;
    }

    public final int getFileId() {
        return _fileId;
    }

    public AbstractTTreeNode getLeft() {
        return _left;
    }

    public Object getMaxKey() throws AxionException {
        return getKey(_nodeSize - 1);
    }

    public Object getMinKey() throws AxionException {
        return getKey(0);
    }

    public AbstractTTreeNode getRight() {
        return _right;
    }

    public final TTree getTree() {
        return _tree;
    }

    // Return value: Has the tree grown in height?
    public final boolean insert(Object key, int value, boolean unique, NodeReference ref) throws AxionException {
        int n = _nodeSize;
        int diff = _tree.compare(key, getKey(0));
        AbstractTTreeNode pg;
        if (diff <= 0) {
            if (diff == 0 && unique) {
                throw new AxionException("Expected the indexed column" + " to be unique, found '" + key + "' already.");
            }
            if ((_left == null || diff == 0) && n != _tree.getMaxNodeSize()) {
                insertKeyValue(0, key, value);
                return false;
            }
            if (_left == null) {
                linkAsLeft(_tree.newNode());
                _left.addKeyValue(key, value);
            } else {
                pg = ref.pg;
                ref.pg = _left;
                boolean grow = _left.insert(key, value, unique, ref);
                if (_left != ref.pg) {
                    _left = ref.pg;
                    markDirty();
                }
                ref.pg = pg;
                if (!grow)
                    return false;
            }
            return balanceRightBranch(ref, true) == 0 ? true : false;
        }
        diff = _tree.compare(key, getKey(n - 1));
        if (diff >= 0) {
            if (diff == 0 && unique) {
                throw new AxionException("Expected the indexed column" + " to be unique, found '" + key + "' already.");
            }
            if ((_right == null || diff == 0) && n != _tree.getMaxNodeSize()) {
                addKeyValue(key, value);
                return false;
            }
            if (_right == null) {
                linkAsRight(_tree.newNode());
                _right.addKeyValue(key, value);
            } else {
                pg = ref.pg;
                ref.pg = _right;
                boolean grow = _right.insert(key, value, unique, ref);
                if (_right != ref.pg) {
                    _right = ref.pg;
                    markDirty();
                }
                ref.pg = pg;
                if (!grow)
                    return false;
            }
            return balanceLeftBranch(ref, true) == 0 ? true : false;
        }
        int l = 1, r = n - 1;
        while (l < r) {
            int i = (l + r) >> 1;
            diff = _tree.compare(key, getKey(i));
            if (diff > 0) {
                l = i + 1;
            } else {
                r = i;
                if (diff == 0) {
                    if (unique) {
                        throw new AxionException("Expected the indexed column" + " to be unique, found '" + key + "' already.");
                    }
                    break;
                }
            }
        }
        // Insert before r
        if (n != _tree.getMaxNodeSize()) {
            insertKeyValue(r, key, value);
            return false;
        } else {
            Object tKey;
            int tValue;
            if (_balance >= 0) {
                tKey = getKey(0);
                tValue = getValue(0);
                copyKeysValues(1, 0, r - 1);
                setKeyValue(r - 1, key, value);
            } else {
                tKey = getKey(n - 1);
                tValue = getValue(n - 1);
                copyKeysValues(r, r + 1, _nodeSize - r - 1);
                setKeyValue(r, key, value);
            }
            return insert(tKey, tValue, unique, ref);
        }
    }

    public abstract void insertKeyValue(int index, Object key, int value) throws AxionException;

    public final boolean isLeaf() {
        return _left == null && _right == null;
    }

    public final boolean isValid() {
        return true;
    }

    public void markDirty() {
        if (!_dirty) {
            _tree.getMetaData().setDirty(this);
            _dirty = true;
        }
    }

    public final void read() throws AxionException {
        ObjectInputStream in = null;
        try {
            File file = _tree.getMetaData().getFileById(_fileId);
            in = FS.openObjectInputSteam(file);

            // Read node size
            _nodeSize = in.readInt();

            if (_tree.isMemoryOnly()) {
                // Load data right away
                Object[] keys = createKeys();
                int[] values = createValues();
                for (int i = 0; i < _nodeSize; i++) {
                    Object key = in.readObject();
                    int value = in.readInt();
                    keys[i] = key;
                    values[i] = value;
                }
            } else {
                // We can load data later
                for (int i = 0; i < _nodeSize; i++) {
                    in.readObject();
                    in.readInt();
                }
            }

            // Read balance
            _balance = in.readInt();
            // Read childs
            int hasLeft = in.readInt();
            if (hasLeft == 1) {
                _left = _tree.newNode(in.readInt());
            } else {
                _left = null;
            }
            int hasRight = in.readInt();
            if (hasRight == 1) {
                _right = _tree.newNode(in.readInt());
            } else {
                _right = null;
            }
        } catch (ClassNotFoundException e) {
            throw new AxionException(e);
        } catch (IOException e) {
            throw new AxionException(e);
        } finally {
            FS.closeInputStream(in);
        }
    }

    public abstract void removeKeyValue(int index) throws AxionException;

    // Slow select.
    // Return value: do we need to continue search?
    public final boolean select(Object minValue, Object maxValue, int inclusive, TTree.Processor proc, boolean first) throws AxionException {
        int l, r, m, n = _nodeSize;

        if (minValue != null && !_tree.isLess(minValue, getMinKey(), inclusive)) {
            if (!_tree.isLess(minValue, getMaxKey(), inclusive)) {
                if (_right != null) {
                    return _right.select(minValue, maxValue, inclusive, proc, first);
                }
                return true;
            }
            for (l = 0, r = n; l < r;) {
                m = (l + r) >> 1;
                if (!_tree.isLess(minValue, getKey(m), inclusive)) {
                    l = m + 1;
                } else {
                    r = m;
                }
            }
            while (r < n) {
                if (maxValue != null && !_tree.isLess(getKey(r), maxValue, inclusive)) {
                    return false;
                }
                proc.process(this, r);
                if (first) {
                    return false;
                }
                r++;
            }
            if (_right != null) {
                return _right.select(minValue, maxValue, inclusive, proc, first);
            }
            return true;
        }
        if (_left != null) {
            if (!_left.select(minValue, maxValue, inclusive, proc, first)) {
                return false;
            }
        }
        for (l = 0; l < n; l++) {
            if (maxValue != null && !_tree.isLess(getKey(l), maxValue, inclusive)) {
                return false;
            }
            proc.process(this, l);
            if (first) {
                return false;
            }
        }
        if (_right != null) {
            return _right.select(minValue, maxValue, inclusive, proc, first);
        }
        return true;
    }

    public abstract void setKeyValue(int index, Object key, int value) throws AxionException;

    public final String toString() {
        return "" + getFileId();
    }

    public final void traverseBackward(TTree.Processor proc) throws AxionException {
        if (_right != null) {
            _right.traverseBackward(proc);
        }
        proc.process(this);
        if (_left != null) {
            _left.traverseBackward(proc);
        }
    }

    public final void traverseForward(TTree.Processor proc) throws AxionException {
        if (_left != null) {
            _left.traverseForward(proc);
        }
        proc.process(this);
        if (_right != null) {
            _right.traverseForward(proc);
        }
    }

    public final void write() throws AxionException {
        ObjectOutputStream out = null;
        try {
            File file = _tree.getMetaData().getFileById(_fileId);
            out = FS.createObjectOutputSteam(file);

            // Write data
            out.writeInt(_nodeSize);
            Object[] keys = getKeys();
            int[] values = getValues();
            for (int i = 0; i < _nodeSize; i++) {
                out.writeObject(keys[i]);
                out.writeInt(values[i]);
            }

            // Write balance
            out.writeInt(_balance);

            // Write childs
            if (_left != null) {
                out.writeInt(1);
                out.writeInt(_left.getFileId());
            } else {
                out.writeInt(0);
            }
            if (_right != null) {
                out.writeInt(1);
                out.writeInt(_right.getFileId());
            } else {
                out.writeInt(0);
            }
        } catch (IOException e) {
            throw new AxionException(e);
        } finally {
            FS.closeOutputStream(out);
        }
    }

    // Return value:
    // 1: The tree changed in height.
    // 0: The tree didn't change in height.
    protected final int balanceLeftBranch(NodeReference ref, boolean insert)
            throws AxionException {
        if (_balance < 0) {
            _balance = 0;
            markDirty();
            return 1;
        } else if (_balance == 0) {
            _balance = 1;
            markDirty();
            return 0;
        } else {
            AbstractTTreeNode righty = this._right;
            if (righty._balance > 0) { // single RR turn (1)
                this._right = righty._left;
                righty._left = this;
                _balance = 0;
                righty._balance = 0;
                ref.pg = righty;
                righty.markDirty();
                markDirty();
                return 1;
            } else if (righty._balance == 0 && !insert) { // single RR turn (2)
                this._right = righty._left;
                righty._left = this;
                _balance = 1;
                righty._balance = -1;
                ref.pg = righty;
                righty.markDirty();
                markDirty();
                return 0;
            } else { // double RL turn
                AbstractTTreeNode lefty = righty._left;
                righty._left = lefty._right;
                lefty._right = righty;
                this._right = lefty._left;
                lefty._left = this;
                _balance = lefty._balance > 0 ? -1 : 0;
                righty._balance = lefty._balance < 0 ? 1 : 0;
                lefty._balance = 0;
                ref.pg = lefty;
                lefty.markDirty();
                righty.markDirty();
                markDirty();
                return 1;
            }
        }
    }

    // Return value:
    // 1: The tree changed in height.
    // 0: The tree didn't change in height.
    protected final int balanceRightBranch(NodeReference ref, boolean insert)
            throws AxionException {
        if (_balance > 0) {
            _balance = 0;
            markDirty();
            return 1;
        } else if (_balance == 0) {
            _balance = -1;
            markDirty();
            return 0;
        } else {
            AbstractTTreeNode lefty = this._left;
            if (lefty._balance < 0) { // single LL turn (1)
                this._left = lefty._right;
                lefty._right = this;
                _balance = 0;
                lefty._balance = 0;
                ref.pg = lefty;
                lefty.markDirty();
                markDirty();
                return 1;
            } else if (lefty._balance == 0 && !insert) { // single LL turn (2)
                this._left = lefty._right;
                lefty._right = this;
                _balance = -1;
                lefty._balance = 1;
                ref.pg = lefty;
                lefty.markDirty();
                markDirty();
                return 0;
            } else { // double LR turn
                AbstractTTreeNode righty = lefty._right;
                lefty._right = righty._left;
                righty._left = lefty;
                this._left = righty._right;
                righty._right = this;
                _balance = righty._balance < 0 ? 1 : 0;
                lefty._balance = righty._balance > 0 ? -1 : 0;
                righty._balance = 0;
                ref.pg = righty;
                lefty.markDirty();
                righty.markDirty();
                markDirty();
                return 1;
            }
        }
    }

    protected void create() {
        createKeys();
        createValues();
    }

    protected abstract Object[] createKeys();

    protected abstract int[] createValues();

    protected void destroy() {
        destroyKeys();
        destroyValues();
    }

    protected abstract void destroyKeys();

    protected abstract void destroyValues();

    protected abstract Object getKey(int index) throws AxionException;

    protected abstract Object[] getKeys() throws AxionException;

    protected AbstractTTreeNode getNext() {
        return _next;
    }

    protected final int getNodeSize() throws AxionException {
        return _nodeSize;
    }

    protected AbstractTTreeNode getPrev() {
        return _prev;
    }

    protected abstract int getValue(int index) throws AxionException;

    protected abstract int[] getValues() throws AxionException;

    protected abstract void setKey(int index, Object key) throws AxionException;

    protected final void setNodeSize(int nodeSize) throws AxionException {
        _nodeSize = nodeSize;
    }

    protected abstract void setValue(int index, int value) throws AxionException;

    /**
     * Print this node, with every line prefixed by N*space spaces.
     */
    protected final String toString(int space) throws AxionException {
        StringBuffer buf = new StringBuffer();
        buf.append(space(space));
        buf.append(getFileId());
        buf.append(" (");
        buf.append("P:" + (_prev != null ? "" + _prev.getFileId() : ""));
        buf.append(" ");
        buf.append("N:" + (_next != null ? "" + _next.getFileId() : ""));
        buf.append(")");
        buf.append(": ");

        Object[] keys = getKeys();
        int[] values = getValues();
        int length = _nodeSize;
        for (int i = 0; i < length; i++) {
            if (i != 0) {
                buf.append(", ");
            }
            buf.append(keys[i] + "/" + values[i]);
        }

        buf.append("\n");
        if (_left != null) {
            buf.append(_left.toString(space + 1));
        } else {
            buf.append(space(space + 1));
            buf.append("null");
            buf.append("\n");
        }
        if (_right != null) {
            buf.append(_right.toString(space + 1));
        } else {
            buf.append(space(space + 1));
            buf.append("null");
            buf.append("\n");
        }
        return buf.toString();
    }

    private void linkAsLeft(AbstractTTreeNode node) {
        _left = node;
        _left._prev = _prev;
        if (_prev != null) {
            _prev._next = _left;
        }
        _left._next = this;
        _prev = _left;
        if (_left._prev == null) {
            _tree.setFirst(_left);
        }
        markDirty();
    }

    private void linkAsRight(AbstractTTreeNode node) {
        _right = node;
        _right._next = _next;
        if (_next != null) {
            _next._prev = _right;
        }
        _right._prev = this;
        _next = _right;
        if (_right._next == null) {
            _tree.setLast(_right);
        }
        markDirty();
    }

    private final void setFileId(int fileId) {
        _fileId = fileId;
    }

    private void unlinkThis() {
        if (_prev != null) {
            _prev._next = _next;
        } else {
            _tree.setFirst(_next);
        }
        if (_next != null) {
            _next._prev = _prev;
        } else {
            _tree.setLast(_prev);
        }
    }

}

class NodeReference {
    AbstractTTreeNode pg;

    NodeReference(AbstractTTreeNode p) {
        pg = p;
    }
}
