/*
 * $Id: DiskTTree.java,v 1.1 2005/06/30 01:14:44 ahimanikya Exp $
 * =======================================================================
 * Copyright (c) 2005 Axion Development Team.  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * 1. Redistributions of source code must retain the above 
 *    copyright notice, this list of conditions and the following 
 *    disclaimer. 
 *   
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 *   
 * 3. The names "Tigris", "Axion", nor the names of its contributors may 
 *    not be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *  
 * 4. Products derived from this software may not be called "Axion", nor 
 *    may "Tigris" or "Axion" appear in their names without specific prior
 *    written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.ext.indexes.ttree;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.ref.SoftReference;
import java.util.Comparator;

import org.axiondb.AxionException;

/**
 * Disk T-Tree implemenation.
 * 
 * @version $Revision: 1.1 $ $Date: 2005/06/30 01:14:44 $
 * @author Pavels Andrejevs
 */
public class DiskTTree extends TTree {

    private class TTreeNode extends AbstractTTreeNode {

        private SoftReference _keysRef;
        private SoftReference _valuesRef;

        public TTreeNode(TTree tree) throws AxionException {
            this(tree, 0);
        }

        public TTreeNode(TTree tree, int fileId) throws AxionException {
            super(tree, fileId);
        }

        public void addKeyValue(Object key, int value) throws AxionException {
            setKey(_nodeSize, key);
            setValue(_nodeSize, value);
            _nodeSize++;
        }

        public void copyKeysValues(int from, int to, int length) throws AxionException {
            markDirty();
            copyKeys(from, to, length);
            copyValues(from, to, length);
        }

        public void insertKeyValue(int index, Object key, int value) throws AxionException {
            markDirty();
            insertKey(index, key);
            insertValue(index, value);
            _nodeSize++;
        }

        public void removeKeyValue(int index) throws AxionException {
            markDirty();
            removeKey(index);
            removeValue(index);
            _nodeSize--;
        }

        public void setKeyValue(int index, Object key, int value) throws AxionException {
            setKey(index, key);
            setValue(index, value);
        }

        protected void copyKeys(int from, int to, int length) throws AxionException {
            Object[] keys = getKeys();
            System.arraycopy(keys, from, keys, to, length);
        }

        protected final void copyValues(int from, int to, int length) throws AxionException {
            int[] values = getValues();
            System.arraycopy(values, from, values, to, length);
        }

        protected Object[] createKeys() {
            Object[] keys = _keysRef == null ? null : (Object[]) _keysRef.get();
            if (keys == null) {
                keys = new Object[_tree.getMaxNodeSize()];
                _keysRef = new SoftReference(keys);
            }
            return keys;
        }

        protected final int[] createValues() {
            int[] values = _valuesRef == null ? null : (int[]) _valuesRef.get();
            if (values == null) {
                values = new int[_tree.getMaxNodeSize()];
                _valuesRef = new SoftReference(values);
            }
            return values;
        }

        protected void destroyKeys() {
            _keysRef = null;
        }

        protected final void destroyValues() {
            _valuesRef = null;
        }

        protected Object getKey(int index) throws AxionException {
            return getKeys()[index];
        }

        protected Object[] getKeys() throws AxionException {
            loadOnNeed();
            return (Object[]) _keysRef.get();
        }

        protected final int getValue(int index) throws AxionException {
            return getValues()[index];
        }

        protected final int[] getValues() throws AxionException {
            loadOnNeed();
            return (int[]) _valuesRef.get();
        }

        protected void insertKey(int index, Object key) throws AxionException {
            Object[] keys = getKeys();
            System.arraycopy(keys, index, keys, index + 1, _nodeSize - index);
            keys[index] = key;
        }

        protected final void insertValue(int index, int value) throws AxionException {
            int[] values = getValues();
            System.arraycopy(values, index, values, index + 1, _nodeSize - index);
            values[index] = value;
        }

        protected final void loadOnNeed() throws AxionException {
            boolean needKeys = needKeys();
            boolean needValues = needValues();
            if (needKeys || needValues) {
                Object[] keys = null;
                int[] values = null;
                if (needKeys) {
                    keys = createKeys();
                }
                if (needValues) {
                    values = createValues();
                }
                // Then check whether the file exists so it can be filled
                File file = _tree.getMetaData().getFileById(_fileId);
                if (file.exists()) {
                    // If it's not, this is a totally new node: nothing to load
                    // then.
                    ObjectInputStream in = null;
                    try {
                        in = FS.openObjectInputSteam(file);
                        int readNodeSize = in.readInt();
                        if (_nodeSize != readNodeSize) {
                            // Sanity check.
                            throw new AxionException("Node size in memory " + "doesn't match that on disk");
                        }
                        if (needKeys && needValues) {
                            for (int i = 0; i < _nodeSize; i++) {
                                Object key = in.readObject();
                                int value = in.readInt();
                                keys[i] = key;
                                values[i] = value;
                            }
                        } else if (needKeys) {
                            for (int i = 0; i < _nodeSize; i++) {
                                Object key = in.readObject();
                                in.readInt();
                                keys[i] = key;
                            }
                        } else {
                            for (int i = 0; i < _nodeSize; i++) {
                                in.readObject();
                                int value = in.readInt();
                                values[i] = value;
                            }
                        }
                    } catch (ClassNotFoundException e) {
                        throw new AxionException(e);
                    } catch (IOException e) {
                        throw new AxionException(e);
                    } finally {
                        FS.closeInputStream(in);
                    }
                }
            }
        }

        // ---------------------------------------------------------------------

        protected boolean needKeys() {
            Object[] keys = _keysRef == null ? null : (Object[]) _keysRef.get();
            return keys == null;
        }

        protected boolean needValues() {
            int[] values = _valuesRef == null ? null : (int[]) _valuesRef.get();
            return values == null;
        }

        protected void removeKey(int index) throws AxionException {
            Object[] keys = getKeys();
            System.arraycopy(keys, index + 1, keys, index, _nodeSize - index - 1);
        }

        protected final void removeValue(int index) throws AxionException {
            int[] values = getValues();
            System.arraycopy(values, index + 1, values, index, _nodeSize - index - 1);
        }

        protected void setKey(int index, Object key) throws AxionException {
            markDirty();
            getKeys()[index] = key;
        }

        protected final void setValue(int index, int value) throws AxionException {
            markDirty();
            getValues()[index] = value;
        }
    }

    public DiskTTree(int maxNodeSize, Comparator comparator, TTreeMetaData metaData) throws AxionException {
        super(maxNodeSize, comparator, metaData);
    }

    public AbstractTTreeNode newNode() throws AxionException {
        return new TTreeNode(this);
    }

    public AbstractTTreeNode newNode(int fileId) throws AxionException {
        return new TTreeNode(this, fileId);
    }

}
