/*
 * $Id: TTree.java,v 1.1 2005/06/30 01:14:44 ahimanikya Exp $
 * =======================================================================
 * Copyright (c) 2005 Axion Development Team.  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * 1. Redistributions of source code must retain the above 
 *    copyright notice, this list of conditions and the following 
 *    disclaimer. 
 *   
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 *   
 * 3. The names "Tigris", "Axion", nor the names of its contributors may 
 *    not be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *  
 * 4. Products derived from this software may not be called "Axion", nor 
 *    may "Tigris" or "Axion" appear in their names without specific prior
 *    written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.ext.indexes.ttree;

import java.io.File;
import java.util.Comparator;
import java.util.Iterator;

import org.apache.commons.collections.primitives.IntListIterator;
import org.axiondb.AxionException;
import org.axiondb.util.IntListIteratorChain;
import org.axiondb.util.NullObject;

/**
 * T-Tree contaniner for objects. T-Tree is most efficient data structure for exact and
 * range data queries in the assumption that all data is present in memory. This class is
 * used in implementation of query iterators to provide fast indexed access to the
 * records.
 * 
 * @version $Revision: 1.1 $ $Date: 2005/06/30 01:14:44 $
 * @author Pavels Andrejevs
 */
public abstract class TTree {

    public abstract static class Processor {
        public void process(AbstractTTreeNode node) throws AxionException {
        }

        public void process(AbstractTTreeNode node, int index) throws AxionException {
        }

        public void process(int value) throws AxionException {
        }
    }

    protected static class NextNodeLinker extends Processor {
        AbstractTTreeNode _next = null;

        public void process(AbstractTTreeNode node) throws AxionException {
            node._next = _next;
            _next = node;
        }
    }

    protected static class NodeChecker extends Processor {
        AbstractTTreeNode first = null;
        AbstractTTreeNode last = null;

        public void process(AbstractTTreeNode node) throws AxionException {
            if (first == null) {
                first = node;
            }
            last = node;
            node.getPrev();
            node.getNext();
        }
    }

    protected static class PrevNodeLinker extends Processor {
        AbstractTTreeNode first = null;
        AbstractTTreeNode last = null;
        AbstractTTreeNode prev = null;

        public void process(AbstractTTreeNode node) throws AxionException {
            node._prev = prev;
            prev = node;
            if (first == null) {
                first = node;
            }
            last = node;
        }
    }

    protected static class TreeForwardNotNullTraverser extends TreeTraverser {
        TreeForwardNotNullTraverser(IntListIteratorChain chain) {
            super(chain);
        }

        public void process(AbstractTTreeNode node) throws AxionException {
            Object nullKey = node.getTree().getNullKey();
            Object[] keys = node.getKeys();
            int[] values = node.getValues();
            int n = node.getNodeSize();
            for (int i = 0; i < n; i++) {
                if (keys[i] != nullKey) {
                    _chain.addIterator(values[i]);
                }
            }
        }
    }

    protected static class TreeForwardNullTraverser extends TreeTraverser {
        TreeForwardNullTraverser(IntListIteratorChain chain) {
            super(chain);
        }

        public void process(AbstractTTreeNode node) throws AxionException {
            Object nullKey = node.getTree().getNullKey();
            Object[] keys = node.getKeys();
            int[] values = node.getValues();
            int n = node.getNodeSize();
            for (int i = 0; i < n; i++) {
                if (keys[i] == nullKey) {
                    _chain.addIterator(values[i]);
                }
            }
        }
    }

    protected static class TreeForwardTraverser extends TreeTraverser {
        TreeForwardTraverser(IntListIteratorChain chain) {
            super(chain);
        }

        public void process(AbstractTTreeNode node) throws AxionException {
            int[] values = node.getValues();
            int n = node.getNodeSize();
            for (int i = 0; i < n; i++) {
                _chain.addIterator(values[i]);
            }
        }
    }

    protected static class TreeSerializer extends Processor {
        java.io.ObjectOutputStream out;

        TreeSerializer(java.io.ObjectOutputStream s) {
            out = s;
        }

        public void process(AbstractTTreeNode node) {
            try {
                out.writeObject(node);
            } catch (java.io.IOException x) {
                throw new Error("Failed to serialize T-Tree: " + x);
            }
        }
    }

    protected static class TreeStripper extends Processor {
        TreeStripper() {
        }

        public void process(AbstractTTreeNode node) throws AxionException {
            node.destroy();
        }
    }

    protected abstract static class TreeTraverser extends Processor {
        IntListIteratorChain _chain;

        TreeTraverser(IntListIteratorChain chain) {
            _chain = chain;
        }

        public void process(AbstractTTreeNode node, int index) throws AxionException {
            _chain.addIterator(node.getValue(index));
        }

        public void process(int value) {
            _chain.addIterator(value);
        }
    }

    protected static class ValueReplacer extends Processor {
        int _newValue;
        int _oldValue;

        ValueReplacer(int oldValue, int newValue) {
            _oldValue = oldValue;
            _newValue = newValue;
        }

        public void process(AbstractTTreeNode node, int index) throws AxionException {
            if (node.getValue(index) == _oldValue) {
                node.setValue(index, _newValue);
            }
        }
    }

    private Comparator _comparator;
    private AbstractTTreeNode _first;
    private AbstractTTreeNode _last;

    private int _maxNodeSize;
    private TTreeMetaData _metaData;
    private AbstractTTreeNode _root;

    // Constructors
    // ---------------------------------------------------------------------

    public TTree(int maxNodeSize, Comparator comparator, TTreeMetaData metaData) throws AxionException {
        _maxNodeSize = maxNodeSize;
        _comparator = comparator;
        _metaData = metaData;
        // Create all nodes
        if (metaData.getFileById(metaData.getRootFileId()).exists()) {
            setRoot(newNode(metaData.getRootFileId()));
            linkNodes();
        } else {
            setRoot(null);
        }
    }

    public final void check() throws AxionException {
        NodeChecker checker = new NodeChecker();
        traverseForward(checker);
        if (checker.first != getFirst()) {
            System.out.println("'FIRST' ERROR");
        }
        if (checker.last != getLast()) {
            System.out.println("'LAST' ERROR");
        }
    }

    public final void delete(Object key, int value) throws AxionException {
        if (getRoot() != null) {
            NodeReference ref = new NodeReference(getRoot());
            getRoot().delete(key, value, ref);
            setRoot(ref.pg);
            size(size() - 1);
        }
        // check();
    }

    public final void deleteAll(Object key) throws AxionException {
        IntListIterator it = getAll(key);
        while (it.hasNext()) {
            int value = it.next();
            delete(key, value);
        }
    }

    public void deleteNode(AbstractTTreeNode node) {
        _metaData.clearDirty(node);
        File file = _metaData.getFileById(node.getFileId());
        if (file.exists()) {
            file.delete();
        }
    }

    public Comparator getComparator() {
        return _comparator;
    }

    public final AbstractTTreeNode getFirst() {
        return _first;
    }

    public final AbstractTTreeNode getLast() {
        return _last;
    }

    public int getMaxNodeSize() {
        return _maxNodeSize;
    }

    public TTreeMetaData getMetaData() {
        return _metaData;
    }

    public int getMinNodeSize() {
        return _maxNodeSize - 1; // if bigger than 1, tests do fail
    }

    public Object getNullKey() {
        return NullObject.INSTANCE;
    }

    public final AbstractTTreeNode getRoot() {
        return _root;
    }

    public final void insert(Object key, int value, boolean unique) throws AxionException {
        if (getRoot() != null) {
            NodeReference ref = new NodeReference(getRoot());
            getRoot().insert(key, value, unique, ref);
            setRoot(ref.pg);
        } else {
            setRoot(newNode());
            getRoot().addKeyValue(key, value);
            setFirst(getRoot());
            setLast(getRoot());
        }
        size(size() + 1);
        // check();
    }
    
    public boolean isMemoryOnly() {
        return _metaData.isMemoryOnly();
    }

    public boolean isValid() {
        return true;
    }

    public abstract AbstractTTreeNode newNode() throws AxionException;

    public abstract AbstractTTreeNode newNode(int fileId) throws AxionException;

    public final void save(File dataDirectory) throws AxionException {
        _metaData.setDataDirectory(dataDirectory);
        if (_metaData.hasDirtyNodes()) {
            _metaData.saveMetaDataFile();
        }
        for (Iterator iter = _metaData.getDirtyNodes(); iter.hasNext();) {
            AbstractTTreeNode node = (AbstractTTreeNode) iter.next();
            node.write();
            node.clearDirty();
            iter.remove();
        }
    }

    public final void saveAfterTruncate(File dataDirectory) throws AxionException {
        _metaData.setAllClean();
        save(dataDirectory);
    }

    public void scanBackward(Processor proc) throws AxionException {
        AbstractTTreeNode node = getLast();
        while (node != null) {
            proc.process(node);
            node = node.getPrev();
        }
    }

    public void scanForward(Processor proc) throws AxionException {
        AbstractTTreeNode node = getFirst();
        while (node != null) {
            proc.process(node);
            node = node.getNext();
        }
    }

    public IntListIteratorChain getInorderIterator() throws AxionException {
        IntListIteratorChain chain = new IntListIteratorChain();
        scanForward(new TreeForwardTraverser(chain));
        return chain;
    }

    public final IntListIterator getAllNotNull() throws AxionException {
        IntListIteratorChain chain = new IntListIteratorChain();
        scanForward(new TreeForwardNotNullTraverser(chain));
        return chain;
    }

    public final IntListIterator getAllNull() throws AxionException {
        IntListIteratorChain chain = new IntListIteratorChain();
        scanForward(new TreeForwardNullTraverser(chain));
        return chain;
    }

    public final void replace(Object key, int oldValue, int newValue) throws AxionException {
        select(key, key, 1, new ValueReplacer(oldValue, newValue), false);
    }

    public final Integer get(Object key) throws AxionException {
        IntListIteratorChain chain = select(key, key, true, true);
        return chain.hasNext() ? new Integer(chain.next()) : null;
    }

    public final IntListIterator getAll(Object key) throws AxionException {
        return select(key, key, true, false);
    }

    public IntListIteratorChain select(Object minValue, Object maxValue, boolean inclusive) throws AxionException {
        return select(minValue, maxValue, inclusive, false);
    }
    
    public IntListIteratorChain select(Object minValue, Object maxValue, boolean inclusive, boolean first) throws AxionException {
        IntListIteratorChain chain = new IntListIteratorChain();
        if (false) {
            if (getRoot() != null) {
                getRoot().select(minValue, maxValue, inclusive ? 1 : 0, new TreeForwardTraverser(chain), first);
            }
        } else {
            select(minValue, maxValue, inclusive ? 1 : 0, new TreeForwardTraverser(chain), first);
        }
        return chain;
    }

    // Ultra-fast
    public final boolean select(Object minValue, Object maxValue, int inclusive, TTree.Processor proc, boolean first) throws AxionException {
        AbstractTTreeNode node = getFirst();
        int r = 0;

        if (minValue != null) {
            AbstractTTreeNode x = getRoot();
            while (x != null) {
                if (isLess(minValue, x.getMinKey(), inclusive)) {
                    x = x.getLeft();
                } else {
                    node = x;
                    x = x.getRight();
                }
            }
            while (node != null && !isLess(minValue, node.getMaxKey(), inclusive)) {
                node = node.getNext();
            }
            if (node == null) {
                return false;
            }
            int l;
            for (l = 0, r = node.getNodeSize(); l < r;) {
                int m = (l + r) >> 1;
                if (!isLess(minValue, node.getKey(m), inclusive)) {
                    l = m + 1;
                } else {
                    r = m;
                }
            }
        }

        if (node == null) {
            node = getFirst();
            if (node == null) {
                return false;
            }
        }

        do {
            int n = node.getNodeSize();
            while (r < n) {
                if (maxValue != null && !isLess(node.getKey(r), maxValue, inclusive)) {
                    return false;
                }
                proc.process(node, r);
                if (first) {
                    return false;
                }
                r++;
            }
            node = node.getNext();
            r = 0;
        } while (node != null);

        return true;
    }

    public final void setFirst(AbstractTTreeNode first) {
        _first = first;
    }

    public final void setLast(AbstractTTreeNode last) {
        _last = last;
    }

    public void setMaxNodeSize(int maxNodeSize) {
        _maxNodeSize = maxNodeSize;
    }

    public final int size() {
        return _metaData.getTreeSize();
    }

    public final String toString() {
        try {
            StringBuffer buf = new StringBuffer();
            buf.append(getRoot() != null ? getRoot().toString(0) : "empty\n");
            buf.append("<");
            buf.append("F:" + (getFirst() != null ? "" + getFirst().getFileId() : ""));
            buf.append(" ");
            buf.append("L:" + (getLast() != null ? "" + getLast().getFileId() : ""));
            buf.append(">\n");
            return buf.toString();
        } catch (AxionException e) {
            return "ERROR: " + e;
        }
    }

    /**
     * Traverse T-Tree elements in descent order
     */
    public void traverseBackward(Processor proc) throws AxionException {
        if (getRoot() != null) {
            getRoot().traverseBackward(proc);
        }
    }

    /**
     * Traverse T-Tree elements in ascent order
     */
    public void traverseForward(Processor proc) throws AxionException {
        if (getRoot() != null) {
            getRoot().traverseForward(proc);
        }
    }

    public void truncate() throws AxionException {
        setRoot(null);
        setFirst(null);
        setLast(null);
    }

    final int compare(Object x, Object y) {
        Comparator c = getComparator();
        Object nullKey = getNullKey();
        if (x == nullKey && y == nullKey) {
            return 0;
        }
        if (x == nullKey && y != nullKey) {
            return -1;
        } else if (y == nullKey && x != nullKey) {
            return 1;
        }
        return c.compare(x, y);
    }

    final boolean isEqual(Object x, Object y) {
        return compare(x, y) == 0;
    }

    final boolean isGreater(Object x, Object y, int inclusive) {
        return compare(x, y) > -inclusive;
    }

    final boolean isGreaterThan(Object x, Object y) {
        return compare(x, y) > 0;
    }

    final boolean isGreaterThanOrEqual(Object x, Object y) {
        return compare(x, y) >= 0;
    }

    final boolean isLess(Object x, Object y, int inclusive) {
        return compare(x, y) < inclusive;
    }

    final boolean isLessThan(Object x, Object y) {
        return compare(x, y) < 0;
    }

    final boolean isLessThanOrEqual(Object x, Object y) {
        return compare(x, y) <= 0;
    }

    final boolean isNotEqual(Object x, Object y) {
        return compare(x, y) != 0;
    }

    protected final void setRoot(AbstractTTreeNode root) {
        _root = root;
        if (root != null) {
            _metaData.setRootFileId(root.getFileId());
        }
    }

    protected final void size(int size) {
        _metaData.setTreeSize(size);
    }

    private void linkNodes() throws AxionException {
        PrevNodeLinker prevLinker = new PrevNodeLinker();
        traverseForward(prevLinker);
        traverseBackward(new NextNodeLinker());
        setFirst(prevLinker.first);
        setLast(prevLinker.last);
    }

}
