/*
 * $Id: TTreeIndex.java,v 1.1 2005/06/30 01:14:44 ahimanikya Exp $
 * =======================================================================
 * Copyright (c) 2005 Axion Development Team.  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * 1. Redistributions of source code must retain the above 
 *    copyright notice, this list of conditions and the following 
 *    disclaimer. 
 *   
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 *   
 * 3. The names "Tigris", "Axion", nor the names of its contributors may 
 *    not be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *  
 * 4. Products derived from this software may not be called "Axion", nor 
 *    may "Tigris" or "Axion" appear in their names without specific prior
 *    written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.ext.indexes.ttree;

import java.io.File;

import org.apache.commons.collections.primitives.ArrayIntList;
import org.apache.commons.collections.primitives.IntCollections;
import org.apache.commons.collections.primitives.IntListIterator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.axiondb.AxionException;
import org.axiondb.Column;
import org.axiondb.Function;
import org.axiondb.Index;
import org.axiondb.IndexLoader;
import org.axiondb.Row;
import org.axiondb.RowIterator;
import org.axiondb.RowSource;
import org.axiondb.Table;
import org.axiondb.engine.indexes.BaseIndex;
import org.axiondb.engine.rowiterators.EmptyRowIterator;
import org.axiondb.engine.rowiterators.LazyRowRowIterator;
import org.axiondb.event.RowEvent;
import org.axiondb.functions.ComparisonFunction;
import org.axiondb.functions.EqualFunction;
import org.axiondb.functions.GreaterThanFunction;
import org.axiondb.functions.GreaterThanOrEqualFunction;
import org.axiondb.functions.IsNotNullFunction;
import org.axiondb.functions.IsNullFunction;
import org.axiondb.functions.LessThanFunction;
import org.axiondb.functions.LessThanOrEqualFunction;

/**
 * Base implementation for T-Tree based {@link Index indices}.
 * 
 * @version $Revision: 1.1 $ $Date: 2005/06/30 01:14:44 $
 * @author Pavels Andrejevs
 */
public abstract class TTreeIndex extends BaseIndex implements Index {

    public static final int DEFAULT_MAXNODESIZE = 100;

    protected static Log log = LogFactory.getLog(TTreeIndex.class);

    protected TTree _tree;

    // ---------------------------------------------------------------------

    public TTreeIndex(String name, Column column, boolean unique) throws AxionException {
        super(name, column, unique);
    }

    public final void changeRowId(Table table, Row row, int oldId, int newId) throws AxionException {
        try {
            int colnum = table.getColumnIndex(getIndexedColumn().getName());
            Object key = row.get(colnum);
            _tree.replace(key, oldId, newId);

        } catch (AxionException e) {
            String msg = "Unable to change row id";
            log.error(msg, e);
            throw new AxionException(msg, e);
        }
    }

    public abstract IndexLoader getIndexLoader();

    public final RowIterator getInorderRowIterator(RowSource source) throws AxionException {
        IntListIterator resultIds = null;
        try {
            resultIds = _tree.getInorderIterator();
        } catch (AxionException e) {
            String msg = "Unable to retrieve values from index" + getName();
            log.error(msg, e);
            throw new AxionException(msg, e);
        }
        return new LazyRowRowIterator(source, resultIds, _tree.size());
    }

    public final RowIterator getRowIterator(RowSource source, Function function, Object value) throws AxionException {

        IntListIterator resultIds = null;
        if (function instanceof ComparisonFunction) {
            Object convertedValue = getIndexedColumn().getDataType().convert(value);

            if (null == convertedValue) {
                // null fails all comparisions I support
                return EmptyRowIterator.INSTANCE;
            }

            if (function instanceof EqualFunction) {
                if (!isUnique()) {
                    resultIds = _tree.getAll(convertedValue);
                } else {
                    Integer result = _tree.get(convertedValue);
                    if (result == null) {
                        resultIds = IntCollections.getEmptyIntListIterator();
                    } else {
                        resultIds = IntCollections.singletonIntListIterator(result.intValue());
                    }
                }
            } else if (function instanceof LessThanFunction) {
                resultIds = _tree.select(null, convertedValue, false);
            } else if (function instanceof LessThanOrEqualFunction) {
                resultIds = _tree.select(null, convertedValue, true);
            } else if (function instanceof GreaterThanFunction) {
                resultIds = _tree.select(convertedValue, null, false);
            } else if (function instanceof GreaterThanOrEqualFunction) {
                resultIds = _tree.select(convertedValue, null, true);
            } else {
                throw new AxionException("Unsupported function " + function);
            }
        } else if (function instanceof IsNotNullFunction) {
            resultIds = _tree.getAllNotNull();
        } else if (function instanceof IsNullFunction) {
            resultIds = _tree.getAllNull();
        } else {
            throw new AxionException("Unsupported function " + function);
        }

        ArrayIntList ids = new ArrayIntList();
        while (resultIds.hasNext()) {
            ids.add(resultIds.next());
        }
        return new LazyRowRowIterator(source, ids.listIterator(), ids.size());
    }

    public TTree getTTree() {
        return _tree;
    }

    public String getType() {
        return Index.TTREE;
    }

    public final void rowDeleted(RowEvent event) throws AxionException {
        String colName = getIndexedColumn().getName();
        int colIndex = event.getTable().getColumnIndex(colName);
        Object key = event.getOldRow().get(colIndex);
        key = (key == null) ? _tree.getNullKey() : key;
        int rowId = event.getOldRow().getIdentifier();

        try {
            _tree.delete(key, rowId);
        } catch (AxionException e) {
            String msg = "Unable to delete from index " + getName();
            log.error(msg, e);
            throw new AxionException(msg, e);
        }
    }

    // TABLE MODIFICATION LISTENER
    // ---------------------------------------------------------------------

    public final void rowInserted(RowEvent event) throws AxionException {
        String colName = getIndexedColumn().getName();
        int colIndex = event.getTable().getColumnIndex(colName);
        Object key = event.getNewRow().get(colIndex);
        int rowId = event.getNewRow().getIdentifier();

        try {
            if (key == null) {
                key = _tree.getNullKey();
                _tree.insert(key, rowId, false);
            } else {
                _tree.insert(key, rowId, isUnique());
            }
        } catch (AxionException e) {
            String msg = "Unable to insert into index " + getName();
            log.error(msg, e);
            throw new AxionException(msg, e);
        }
    }

    public final void rowUpdated(RowEvent event) throws AxionException {
        rowDeleted(event);
        rowInserted(event);
    }

    public void save(File dataDirectory) throws AxionException {
        getIndexLoader().saveIndex(this, dataDirectory);
    }

    public void saveAfterTruncate(File dataDirectory) throws AxionException {
        getIndexLoader().saveIndexAfterTruncate(this, dataDirectory);
    }

    public boolean supportsFunction(Function fn) {
        if (fn instanceof EqualFunction) {
            return true;
        } else if (fn instanceof LessThanFunction) {
            return true;
        } else if (fn instanceof LessThanOrEqualFunction) {
            return true;
        } else if (fn instanceof GreaterThanFunction) {
            return true;
        } else if (fn instanceof GreaterThanOrEqualFunction) {
            return true;
        } else if (fn instanceof IsNotNullFunction) {
            return true;
        } else if (fn instanceof IsNullFunction) {
            return true;
        } else {
            return false;
        }
    }

    public void truncate() throws AxionException {
        _tree.truncate();
    }

}
