/*
 * $Id: AbstractTestTTreeInt.java,v 1.2 2007/11/13 19:04:02 rwald Exp $
 * =======================================================================
 * Copyright (c) 2002 Axion Development Team.  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * 1. Redistributions of source code must retain the above 
 *    copyright notice, this list of conditions and the following 
 *    disclaimer. 
 *   
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 *   
 * 3. The names "Tigris", "Axion", nor the names of its contributors may 
 *    not be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *  
 * 4. Products derived from this software may not be called "Axion", nor 
 *    may "Tigris" or "Axion" appear in their names without specific prior
 *    written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.ext.indexes.ttree;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.apache.commons.collections.comparators.ComparableComparator;
import org.apache.commons.collections.primitives.ArrayIntList;
import org.apache.commons.collections.primitives.IntIterator;
import org.apache.commons.collections.primitives.IntList;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.axiondb.AbstractDbdirTest;
import org.axiondb.ext.indexes.ttree.TTree;
import org.axiondb.ext.indexes.ttree.DiskTTree;
import org.axiondb.ext.indexes.ttree.TTreeMetaData;

/**
 * @version $Revision: 1.2 $ $Date: 2007/11/13 19:04:02 $
 * @author Pauls Andrejevs
 */
public class AbstractTestTTreeInt extends AbstractDbdirTest {
    private static Log _log = LogFactory.getLog(AbstractTestTTreeInt.class);
    private TTree _tree = null;
    private int _deg = 2;
    private int _max = (int) Math.pow(2 * _deg, 2);
    private Random _random = new Random();
    private static Integer[] _intArray = { 
        new Integer(12345), new Integer(22131), new Integer(33756), new Integer(67890), 
        new Integer(77838), new Integer(88183), new Integer(99312), new Integer(111037), 
        new Integer(222338), new Integer(333192), new Integer(444938), new Integer(555657), 
        new Integer(661027), new Integer(778046), new Integer(889282), new Integer(942739), 
        new Integer(1190642), new Integer(2222567), new Integer(3333643), new Integer(4490833), 
        new Integer(5555677), new Integer(6661278), new Integer(7797355), new Integer(8832121)};
    private TTreeMetaData _metaData = null; 

    //------------------------------------------------------------ Conventional

    public AbstractTestTTreeInt(String testName) {
        super(testName);
    }

    public void setMetaData(TTreeMetaData metaData) {
        this._metaData = metaData;
    }

    public TTreeMetaData getMetaData() {
        return this._metaData;
    }

    //--------------------------------------------------------------- Lifecycle

    public void setUp() throws Exception {
        super.setUp();
        long newSeed = _random.nextLong();
        _log.debug("New Seed: " + newSeed);
        _random.setSeed(newSeed);
        setMetaData(new TTreeMetaData("idx", getDbdir(), getDbdir() == null));
        _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
    }

    public void tearDown() throws Exception {
        _tree = null;
        super.tearDown();
    }

    //------------------------------------------------------------------- Tests

    public void testCreate() throws Exception {
        assertTrue("Should have dbdir avail", getDbdir().exists());
        assertTrue("Dbdir should be directory", getDbdir().isDirectory());
        assertNotNull("Should not be null", _tree);
        assertTrue("Should be empty", _tree.getRoot() == null);
        assertEquals("Should have no entries", 0, _tree.size());
        File idx = new File(getDbdir(), "IDX" + ".1");
        assertTrue("Should not have an index file before save", !idx.exists());
        _tree.insert(new Integer(12345), 1, false);
        _tree.save(getDbdir());
        assertTrue("Should have an index file after save", idx.exists());
    }

    public void testInsertAndRetrieve() throws Exception {
        _tree.insert(new Integer(67890), 456, false);
        assertTrue("Tree should be valid after insert", _tree.isValid());
        assertEquals("Should have one value", 1, _tree.size());
        assertEquals("Should get value back", new Integer(456), _tree.get(new Integer(67890)));
    }

    public void testInsertAndRetrieveMany() throws Exception {
        for (int i = 0; i < _intArray.length; i++) {
            _log.debug("Inserting key " + i);
            _tree.insert(_intArray[i], _intArray.length - i, false);
            _log.debug(_tree.toString());
            assertTrue("Tree should be valid after insert", _tree.isValid());
            assertEquals("Should get the value back as inserting for " + i, new Integer(
                _intArray.length - i), _tree.get(_intArray[i]));
        }
        for (int i = 0; i < _intArray.length; i++) {
            assertEquals("Should get the value back for " + i,
                new Integer(_intArray.length - i), _tree.get(_intArray[i]));
        }
    }

    public void testReloadTree() throws Exception {
        //Insert a bunch of items
        {
            for (int i = 0; i < _intArray.length; i++) {
                _log.debug("Inserting key " + i);
                _tree.insert(_intArray[i], _intArray.length - i, false);
                _log.debug(_tree.toString());
                assertTrue("Tree should be valid after insert", _tree.isValid());
                assertEquals("Should get the value back as inserting for [" + i + "]", 
                        new Integer(_intArray.length - i), _tree.get(_intArray[i]));
            }
            _tree.save(getDbdir());
        }

        //Reload tree and make sure items are still there
        //then delete some of them
        {
            _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
            for (int i = 0; i < _intArray.length; i++) {
                assertEquals("Should get the value back for " + i, new Integer(_intArray.length
                    - i), _tree.get(_intArray[i]));
            }

            for (int i = 0; i < _intArray.length / 2; i++) {
                _tree.delete(_intArray[i], _intArray.length - i);
                assertEquals("Should not get the value back for " + i, null, _tree
                    .get(_intArray[i]));
            }
            _tree.save(getDbdir());
        }

        //Reload the tree and make sure that items that should be there are there
        //and those that should not aren't. Then add the deleted items back
        {
            _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
            for (int i = _intArray.length / 2; i < _intArray.length; i++) {
                assertEquals("Should get the value back for " + i, new Integer(_intArray.length
                    - i), _tree.get(_intArray[i]));
            }

            for (int i = 0; i < _intArray.length / 2; i++) {
                assertEquals("Should not get the value back for " + i, null, _tree
                    .get(_intArray[i]));
                _tree.insert(_intArray[i], _intArray.length - i, false);
            }
            _tree.save(getDbdir());
        }

        //Reload the tree and make sure that all items are there
        {
            _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
            for (int i = 0; i < _intArray.length; i++) {
                assertEquals("Should get the value back for " + i, new Integer(_intArray.length
                    - i), _tree.get(_intArray[i]));
            }
            _tree.save(getDbdir());
        }
    }

    public void testInsertIdenticalAndGetAll() throws Exception {
        _tree = new DiskTTree( 8, new ComparableComparator(), getMetaData());
        int max = 100;
        for (int i = 0; i < max; i++) {
            _log.debug("Inserting key " + i);
            _tree.insert(new Integer(i), i, false);
            _tree.insert(new Integer(i), max - i, false);
            _tree.insert(new Integer(i), -1 * i, false);
            _log.debug(_tree.toString());
            assertTrue("Tree should be valid after insert", _tree.isValid());
            IntIterator iter = _tree.getAll(new Integer(i));
            List<Integer> valList = new ArrayList<Integer>();
            valList.add(new Integer(i));
            valList.add(new Integer(max - i));
            valList.add(new Integer(-1 * i));
            for (int j = 0; j < 3; j++) {
                assertTrue("Should have another value for: " + i, iter.hasNext());
                Integer val = new Integer(iter.next());
                assertTrue("Should contain value: " + val + " for: " + i, valList.contains(val));
                valList.remove(val);
            }
        }
        for (int i = 0; i < max; i++) {
            IntIterator iter = _tree.getAll(new Integer(i));
            List<Integer> valList = new ArrayList<Integer>();
            valList.add(new Integer(i));
            valList.add(new Integer(max - i));
            valList.add(new Integer(-1 * i));
            for (int j = 0; j < 3; j++) {
                assertTrue("Should have another value for: " + i, iter.hasNext());
                Integer val = new Integer(iter.next());
                assertTrue("Should contain value: " + val + " for: " + i, valList.contains(val));
                valList.remove(val);
            }
        }
    }

    public void testInsertAndRetrieveManyRandom() throws Exception {
        Map tests = new HashMap();
        for (int i = 0; i < _intArray.length; i++) {
            Integer key;
            while (tests.containsKey(key = new Integer(_random.nextInt(_intArray.length)))) {
            };
            Integer value = new Integer(_random.nextInt());
            if (!tests.containsKey(key)) {
                tests.put(key, value);
                _log.debug("Inserting key " + key);
                _tree.insert(_intArray[key.intValue()], value.intValue(), false);
                _log.debug(_tree.toString());
                assertTrue("Tree should be valid after insert", _tree.isValid());
                assertEquals("Should get value back as inserting", value, _tree
                    .get(_intArray[key.intValue()]));
            }
        }
        assertEquals("Should have " + _intArray.length + " insertions", _intArray.length,
            tests.keySet().size());
        Iterator i = tests.keySet().iterator();
        while (i.hasNext()) {
            Integer cur = (Integer) i.next();
            assertEquals("Should get value back", tests.get(cur), _tree.get(_intArray[cur
                .intValue()]));
        }
    }

    public void testInsertAndRetrieveManyIdentical() throws Exception {
        Set tests = new HashSet();
        for (int i = 0; i < _max; i++) {
            _log.debug("Insertion number " + (i + 1));
            _tree.insert(new Integer(12345), i, false);
            assertTrue("Tree should be valid after insert", _tree.isValid());
            tests.add(new Integer(i));
            _log.debug(_tree.toString());
        }
        assertEquals("Should have full number of tests", _max, tests.size());
        IntIterator i = _tree.getAll(new Integer(12345));
        int count = 0;
        while (i.hasNext()) {
            tests.remove(new Integer(i.next()));
            count++;
        }
        assertEquals("Should have found all matching inserts", _max, count);
        assertTrue("All test vals should have been retrieved", tests.isEmpty());
        i = _tree.getAll(new Integer(67890));
        assertTrue("Should not have any matches", !i.hasNext());
    }

    public void testInsertAndDeleteLots() throws Exception {
        //Do inserts
        for (int i = 0; i < _intArray.length; i++) {
            _log.debug("Inserting key " + i);
            _tree.insert(_intArray[i], i, false);
            _log.debug(_tree.toString());
            assertTrue("Tree should be valid after insert", _tree.isValid());
            assertEquals("Should get value back as inserting", i, _tree.get(_intArray[i])
                .intValue());
        }

        for (int i = 0; i < _intArray.length; i++) {
            int val = (i + 20) % (_intArray.length);
            assertEquals("Should get value back", val, _tree.get(_intArray[val]).intValue());
            _log.debug("Deleting key " + val);
            _log.debug("Tree before delete: " + _tree.toString());
            _tree.delete(_intArray[val], val);
            _log.debug("Tree after delete: " + _tree.toString());
            assertTrue("Tree should be valid after delete", _tree.isValid());
            assertNull("Value should be deleted for key: " + val, _tree.get(_intArray[val]));
        }

        assertTrue("Tree should be empty", _tree.size() == 0);
    }

    public void testInsertAndDeleteLotsWithReload() throws Exception {
        //Do inserts
        for (int i = 0; i < _intArray.length; i++) {
            _log.debug("Inserting key " + i);
            _tree.insert(_intArray[i], i, false);
            _log.debug(_tree.toString());
            assertTrue("Tree should be valid after insert", _tree.isValid());
            assertEquals("Should get value back as inserting", i, _tree.get(_intArray[i])
                .intValue());
        }
        _tree.save(getDbdir());

        _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
        for (int i = 0; i < _intArray.length; i++) {
            int val = (i + 20) % (_intArray.length);
            assertEquals("Should get value back", val, _tree.get(_intArray[val]).intValue());
            _log.debug("Deleting key " + val);
            _log.debug("Tree before delete: " + _tree.toString());
            _tree.delete(_intArray[val], val);
            _log.debug("Tree after delete: " + _tree.toString());
            assertTrue("Tree should be valid after delete", _tree.isValid());
            assertNull("Value should be deleted for key: " + val, _tree.get(_intArray[val]));
        }
        _tree.save(getDbdir());

        _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
        assertTrue("Tree should be empty", _tree.size() == 0);
    }

    public void testInsertAndDeleteRandom() throws Exception {
        Map tests = new HashMap();
        //Do inserts
        int max = _max * 10;
        for (int i = 0; i < max; i++) {
            Integer value = new Integer(_random.nextInt());
            Integer key = value;
            if (!tests.containsKey(key)) {
                tests.put(key, value);
                _log.debug("Inserting key " + key);
                _tree.insert(key, value.intValue(), false);
                _log.debug(_tree.toString());
                assertTrue("Tree should be valid after insert", _tree.isValid());
                assertEquals("Should get value back as inserting", value, _tree.get(key));
            }
        }
        assertEquals("Should have " + max + " insertions", max, tests.keySet().size());
        //Do deletes
        Iterator i = tests.keySet().iterator();
        while (i.hasNext()) {
            Integer cur = (Integer) i.next();
            assertEquals("Should get value back", tests.get(cur), _tree.get(cur));
            _log.debug("Deleting key " + cur);
            _log.debug("Tree before delete: " + _tree.toString());
            _tree.delete(cur, ((Integer)tests.get(cur)).intValue());
            _log.debug("Tree after delete: " + _tree.toString());
            assertTrue("Tree should be valid after delete", _tree.isValid());
            assertNull("Value should be deleted for key: " + cur, _tree.get(cur));
        }
        assertEquals("Tree should be empty", 0, _tree.size());
    }

    public void testInsertAndDeleteRandomWithReload() throws Exception {
        Map tests = new HashMap();
        //Do inserts
        int max = _max * 10;
        for (int i = 0; i < max; i++) {
            Integer value = new Integer(_random.nextInt());
            Integer key = value;
            if (!tests.containsKey(key)) {
                tests.put(key, value);
                _log.debug("Inserting key " + key);
                _tree.insert(key, value.intValue(), false);
                _log.debug(_tree.toString());
                assertTrue("Tree should be valid after insert", _tree.isValid());
                assertEquals("Should get value back as inserting", value, _tree.get(key));
            }
        }
        assertEquals("Should have " + max + " insertions", max, tests.keySet().size());

        _tree.save(getDbdir());

        _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
        //Do deletes
        Iterator i = tests.keySet().iterator();
        while (i.hasNext()) {
            Integer cur = (Integer) i.next();
            assertEquals("Should get value back", tests.get(cur), _tree.get(cur));
            _log.debug("Deleting key " + cur);
            _log.debug("Tree before delete: " + _tree.toString());
            _tree.delete(cur, ((Integer)tests.get(cur)).intValue());
            _log.debug("Tree after delete: " + _tree.toString());
            assertTrue("Tree should be valid after delete", _tree.isValid());
            assertNull("Value should be deleted for key: " + cur, _tree.get(cur));
        }

        _tree.save(getDbdir());

        _tree = new DiskTTree( _deg, new ComparableComparator(), getMetaData());
        assertEquals("Tree should be empty", 0, _tree.size());
    }

    public void testGetAllTo() throws Exception {
        testInsertAndRetrieveMany();

        int topNum = 19;
        Integer topKey = _intArray[topNum];
        IntIterator iter = _tree.select(null, topKey, true);
        for (int i = 0; i < topNum; i++) {
            assertTrue("Iterator should have another value", iter.hasNext());
            Integer val = new Integer(iter.next());
            _log.debug("Iterator Value: " + val);
            assertEquals("Should get the value back for " + i,
                new Integer(_intArray.length - i), val);
        }
    }

    public void testGetAllFrom() throws Exception {
        testInsertAndRetrieveMany();

        int bottomNum = 19;
        Integer bottomKey = _intArray[bottomNum];
        IntIterator iter = _tree.select(bottomKey, null, true);
        for (int i = bottomNum; i < _intArray.length; i++) {
            assertTrue("Iterator should have another value", iter.hasNext());
            Integer val = new Integer(iter.next());
            _log.debug("Iterator Value: " + val);
            assertEquals("Should get the value back for " + i,
                new Integer(_intArray.length - i), val);
        }
    }

    public void testChangeId() throws Exception {
        testInsertAndRetrieveMany();

        Integer key = new Integer(67890);
        int oldId = (_tree.get(key)).intValue();
        _tree.replace(key, oldId, oldId + 1000);
        assertEquals("Should get new id back for " + key, oldId + 1000, ( _tree.get(key))
            .intValue());
    }

    public void testChangeIdIdentical() throws Exception {
        testInsertAndRetrieveManyIdentical();

        Integer key = new Integer(12345);
        IntList oldIdList = new ArrayIntList();
        IntIterator iter = null;
        {
            ArrayIntList ids = new ArrayIntList();
            for (IntIterator tocopy = _tree.getAll(key); tocopy.hasNext();) {
                ids.add(tocopy.next());
            }
            iter = ids.iterator();
        }
        for (int i = 0; iter.hasNext(); i++) {
            int oldId = iter.next();
            oldIdList.add(oldId);
            if ((i % 2) == 0) {
                _tree.replace(key, oldId, oldId + 1000);
            }
        }
        {
            ArrayIntList ids = new ArrayIntList();
            for (IntIterator tocopy = _tree.getAll(key); tocopy.hasNext();) {
                ids.add(tocopy.next());
            }
            iter = ids.iterator();
        }
        for (int i = 0; iter.hasNext(); i++) {
            int newId = iter.next();
            int oldId = oldIdList.get(i);
            if ((i % 2) == 0) {
                assertEquals("Should get new id back for " + key, oldId + 1000, newId);
            } else {
                assertEquals("Should get old id back for " + key, oldId, newId);
            }
        }
    }
}