/*
 * $Id: DelimitedFlatfileTable.java,v 1.18 2006/01/10 21:02:36 ahimanikya Exp $
 * =======================================================================
 * Copyright (c) 2002-2004 Axion Development Team.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The names "Tigris", "Axion", nor the names of its contributors may
 *    not be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. Products derived from this software may not be called "Axion", nor
 *    may "Tigris" or "Axion" appear in their names without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 */

package org.axiondb.engine.tables;

import java.io.CharArrayWriter;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.regex.Pattern;

import org.axiondb.AxionException;
import org.axiondb.DataType;
import org.axiondb.Database;
import org.axiondb.ExternalTable;
import org.axiondb.Row;
import org.axiondb.engine.rows.SimpleRow;
import org.axiondb.io.BufferedDataInputStream;
import org.axiondb.io.BufferedDataOutputStream;

/**
 * A disk-resident Delimited Flatfile {@link org.axiondb.Table}.<br>
 * 
 * TODO: Support for multiple delimiter for field and record
 * TODO: Support for treating consecutive delimiter as one
 * 
 * @version $Revision: 1.18 $ $Date: 2006/01/10 21:02:36 $
 * @author Ahimanikya Satapathy
 * @author Jonathan Giron
 */
public class DelimitedFlatfileTable extends BaseFlatfileTable {

    private static final String EMPTY_STRING = "";
    private static final String COMMA = ",";
    private static final char NL = Character.MAX_VALUE;
    public static final String PROP_FIELDDELIMITER = "FIELDDELIMITER"; // NOI18N
    public static final String PROP_QUALIFIER = "QUALIFIER"; // NOI18N
    
    byte[] QUALIFIER_BYTES;
    byte[] EMPTY_STRING_BYTES = EMPTY_STRING.getBytes();
    byte[] LINESEP_BYTES;
    byte[] FIELDSEP_BYTES;

    private static final Set PROPERTY_KEYS = new HashSet(2);
    
    /** Set of required keys for organization properties */
    private static final Set REQUIRED_KEYS = new HashSet(1);

    static {
        PROPERTY_KEYS.add(PROP_FIELDDELIMITER);
        PROPERTY_KEYS.add(PROP_QUALIFIER);
    }

    public DelimitedFlatfileTable(String name, Database db) throws AxionException {
        super(name, db, new DelimitedFlatfileTableLoader());
        setType(ExternalTable.DELIMITED_TABLE_TYPE);
    }

    protected String getDefaultDataFileExtension() {
        return "csv";
    }

    protected String getQualifier() {
        return _qualifier;
    }

    protected int getQualifierLength() {
        return _qualifier.length();
    }

    protected Row getRowByOffset(int idToAssign, long ptr) throws AxionException {
        BufferedDataInputStream data = getInputStream();
        int colCount = getColumnCount();
        Row row = new SimpleRow(idToAssign, colCount);

        try {
            synchronized (data) {
                char[] charArray = readLine(data, ptr);
                if (charArray[0] == NL) {
                    throw new AxionException("Empty line detected - invalid.");
                }
                
                CharTokenizer charTokenizer = new CharTokenizer(charArray, _fieldSep);
    
                for (int i = 0; i < colCount && charTokenizer.hasMoreTokens(); i++) {
                    String columnValue = charTokenizer.nextToken();
                    row = trySettingColumn(idToAssign, row, i, columnValue);
                }
            }
        } catch (Exception e) {
            if (e instanceof AxionException) {
                throw (AxionException) e;
            }
            throw new AxionException(e);
        }
        return row;
    }

    protected boolean isQuoted() {
        return !isNullString(_qualifier);
    }

    protected boolean isEndOfRecord(int recLength, int nextChar, BufferedDataInputStream data) throws IOException {
        if (isEOF(nextChar)) {
            return true;
        }

        boolean foundEOL = false;
        for(int k = 0; (k < _lineSeps.length && !foundEOL) ; k++) {
            String lineSep = _lineSeps[k];
            if (!("".equals(lineSep)) && lineSep.charAt(0) == nextChar) {
                foundEOL = true;
                char[] charBuf = lineSep.toCharArray();
                // Look ahead to see whether the following chars match EOL.
                long lastDataFileOffset = data.getPos();
                for (int i = 1, I = lineSep.length(); i < I; i++) {
                    if (charBuf[i] != (char) data.read()) {
                        data.seek(lastDataFileOffset);
                        foundEOL = false;
                    }
                }
            }
        }
        return foundEOL;
    }

    public boolean loadExternalTable(Properties props) throws AxionException {
        context = new DelimitedTableOrganizationContext();
        return super.loadExternalTable(props);
    }

    public Properties getTableProperties() {
        return context.getTableProperties();
    }

    protected void parseTableProperties(ObjectInputStream in) throws AxionException {
        try {
            _lineSep = in.readUTF();
            _fieldSep = in.readUTF();
            _isFirstLineHeader = Boolean.valueOf(in.readUTF()).booleanValue();
            _fileName = in.readUTF();
            in.readUTF(); // _eol will be computed, keep for older version metadata
            _qualifier = in.readUTF();
            in.readUTF(); // _quoted will be computed, keep for older version metadata
            
            try {
                _rowCount = in.readInt();
            } catch (EOFException ignore) {
                // Goes here if metadata from an older version is parsed - ignore.
            }

            context = new DelimitedTableOrganizationContext();
            context.updateProperties();
            context.readOrSetDefaultProperties(context.getTableProperties());
            createOrLoadDataFile();

        } catch (IOException ioex) {
            throw new AxionException("Unable to parse meta file for table " + getName(), ioex);
        }
    }
    
    protected void writeHeader(BufferedDataOutputStream dataFile) throws AxionException {
        if (_isFirstLineHeader) {
            try {
                CharArrayWriter header = new CharArrayWriter();
                for (int i = 0, I = getColumnCount(); i < I; i++) {

                    if (i != 0) {
                        header.write(_fieldSep);
                    }
                    header.write(getColumn(i).getName());
                }
                header.write(_preferredLineSep);
                dataFile.write(header.toString().getBytes());
                header.close();
            } catch (IOException ioex) {
                throw new AxionException("Unable to write header for table: " + getName(), ioex);
            }
        }
    }

    protected void writeRow(BufferedDataOutputStream buffer, Row row) throws AxionException {
        Object colValue = null;
        DataType type = null;

        QUALIFIER_BYTES = _qualifier.getBytes();
        LINESEP_BYTES = _preferredLineSep.getBytes();
        FIELDSEP_BYTES = _fieldSep.getBytes();
        
        try {
            for (int i = 0, I = getColumnCount(); i < I; i++) {
                colValue = row.get(i);
                type = getColumn(i).getDataType();

                if (i != 0) {
                    buffer.write(FIELDSEP_BYTES);
                }
                byte[] qualifier = isEscapeRequired(type) ? QUALIFIER_BYTES : EMPTY_STRING_BYTES;
                if (colValue != null) {
                    buffer.write(qualifier);
                    String val = type.toString(colValue);
                    if(isQuoted() && val.indexOf(_qualifier) != -1) {
                        // escape the quealifier in the data string.
                        val = _qPattern.matcher(val).replaceAll(_qualifier + _qualifier);
                    }
                    buffer.write(val.getBytes());
                    buffer.write(qualifier);
                } else {
                    buffer.write(EMPTY_STRING_BYTES); // Write Null column
                }
            }
            // write new line
            buffer.write(LINESEP_BYTES);
        } catch (IOException e) {
            throw new AxionException("Error writing row: " + row, e);
        }
    }

    protected void writeTableProperties(ObjectOutputStream out) throws AxionException {
        try {
            if (_lineSep != null && _fieldSep != null && _fileName != null) {
                out.writeUTF(_lineSep);
                out.writeUTF(_fieldSep);
                out.writeUTF(Boolean.toString(_isFirstLineHeader));
                out.writeUTF(_fileName);
                out.writeUTF(_lineSep);
                out.writeUTF(_qualifier);
                out.writeUTF("true");
                out.writeInt(_rowsToSkip);
            }
        } catch (IOException ioex) {
            throw new AxionException("Unable to write meta file for table " + getName(), ioex);
        }
    }

    private boolean isEscapeRequired(DataType type) {
        switch (type.getJdbcType()) {
            case Types.CHAR:
            case Types.DATE:
            case Types.TIME:
            case Types.TIMESTAMP:
            case Types.VARCHAR:
                return true;
            default:
                return false;
        }
    }

    private char[] readLine(BufferedDataInputStream data, long fileOffset) throws AxionException {
        Arrays.fill(_lineCharArray, FILLER);
        int recLength = 0;
        try {
            int nextChar;
            data.seek(fileOffset);

            while (true) {
                nextChar = data.read();
                if (isEndOfRecord(recLength, nextChar, data)) {
                    _lineCharArray[recLength] = NL;
                    break;
                }

                // ensure capacity
                if ((recLength + 2) > _lineCharArray.length) {
                    char[] newlineCharArray = new char[recLength + 80];
                    System.arraycopy(_lineCharArray, 0, newlineCharArray, 0, _lineCharArray.length);
                    _lineCharArray = newlineCharArray;
                }

                _lineCharArray[recLength++] = ((char) nextChar);
            }
            return _lineCharArray;

        } catch (IOException e) {
            throw new AxionException("Unable to parse data file: ", e);
        }
    }

    protected boolean isNewLine(int nextChar) {
        return nextChar == NL;
    }

    private class DelimitedTableOrganizationContext extends BaseFlatfileTableOrganizationContext {
        public Set getPropertyKeys() {
            Set baseKeys = super.getPropertyKeys();
            Set keys = new HashSet(baseKeys.size() + PROPERTY_KEYS.size());
            keys.addAll(baseKeys);
            keys.addAll(PROPERTY_KEYS);

            return keys;
        }

        public void readOrSetDefaultProperties(Properties props) throws AxionException {
            super.readOrSetDefaultProperties(props);

            String rawFieldSep = props.getProperty(PROP_FIELDDELIMITER);
            if (isNullString(rawFieldSep)) {
                rawFieldSep = COMMA;
            }
            _fieldSep = fixEscapeSequence(rawFieldSep);
            
            // default line separator is new line
            String lineSep = System.getProperty("line.separator");
            if("".equals(_lineSep)) {
                _lineSep = fixEscapeSequence(lineSep);
            }
            
            // Support multiple record delimiter for delimited 
            StringTokenizer tokenizer = new StringTokenizer(_lineSep, " ");
            ArrayList tmpList = new ArrayList();
            while(tokenizer.hasMoreTokens()) {
                String token = tokenizer.nextToken();
                tmpList.add(token);
                if(token.equals(lineSep)) {
                    _preferredLineSep = token;
                }
            }
            _lineSeps = (String[])tmpList.toArray(new String[0]);
            
            // determine the delimiter to be used for writing line
            if(_preferredLineSep == null || _preferredLineSep.length() == 0) {
                _preferredLineSep = _lineSeps[0];
            }

            _qualifier = fixEscapeSequence(props.getProperty(PROP_QUALIFIER));
            if (isNullString(_qualifier)) {
                _qualifier = EMPTY_STRING;
            } else {
                _qPattern = Pattern.compile(_qualifier);
                _qqPattern = Pattern.compile(_qualifier + _qualifier);
            }
            
        }

        public void updateProperties() {
            super.updateProperties();

            _props.setProperty(PROP_LOADTYPE, ExternalTableFactory.TYPE_DELIMITED);
            _props.setProperty(PROP_FIELDDELIMITER, addEscapeSequence(_fieldSep));
            _props.setProperty(PROP_QUALIFIER, _qualifier);
        }
        
        public Set getRequiredPropertyKeys() {
            Set baseRequiredKeys = getBaseRequiredPropertyKeys();
            Set keys = new HashSet(baseRequiredKeys.size() + REQUIRED_KEYS.size());
            keys.addAll(baseRequiredKeys);
            keys.addAll(REQUIRED_KEYS);

            return keys;
        }
    }

    class CharTokenizer {
        char[] _charArray;

        private int _currentPosition;

        private String _delimiters;
        private int _maxPosition;

        public CharTokenizer(char[] thecharArray, String theDelim) {
            _delimiters = theDelim;
            _charArray = thecharArray;
            _maxPosition = _charArray.length;
            _currentPosition = 0;
        }

        public boolean hasMoreTokens() {
            return (_currentPosition < _maxPosition);
        }

        public String nextToken() {
            int start = _currentPosition;
            int end = start;
            int pos = _currentPosition;
            boolean inQuotedString = false;
            boolean endQuotedString = false;
            boolean treatAsUnquoted = false;
            boolean wasEscaped = false;

            while (pos < _maxPosition) {
                // if new line
                if (isNewLine(_charArray[pos])) {
                    if (isQuoted() && !endQuotedString) {
                        _maxPosition = pos;
                        _currentPosition = pos;
                        end = pos;
                        break;
                    }
                    _currentPosition = _maxPosition;
                }

                // if quoted and found qualifier
                if (isQuoted() && isQualifier(pos)) {
                    if (!inQuotedString) { // not inside the quoted string
                        pos += getQualifierLength();
                        start = pos;
                        inQuotedString = true;
                        continue;
                    } else if (isQualifier(pos + getQualifierLength())){
                        pos += (getQualifierLength() * 2);
                        wasEscaped = true;
                        continue;
                    }
                    // inside the quoted string
                    end = pos;
                    pos += getQualifierLength();
                    inQuotedString = false;
                    endQuotedString = true;
                    continue;
                }

                // if quoted, close quote found, but have not found a delimiter yet
                if (isQuoted() && endQuotedString && _delimiters.charAt(0) != _charArray[pos]
                    && !isNewLine(_charArray[pos])) {
                    pos++;
                    continue;
                }

                // if quoted, close quote found and found a delimiter
                if (isQuoted() && endQuotedString) {
                    if (isDelimiter(pos)) {
                        pos += _delimiters.length();
                        break;
                    } else if (isNewLine(_charArray[pos])) {
                        break;
                    } 
                }

                // if quoted but did not find start qualifer, treat this token as
                // unquoted
                if (isQuoted() && !inQuotedString) {
                    treatAsUnquoted = true;
                }

                // if non-quoted
                if ((!isQuoted() || treatAsUnquoted) && pos < _maxPosition) {
                    if (isDelimiter(pos)) {
                        end = pos;
                        pos += _delimiters.length();
                        break;
                    } else if (isNewLine(_charArray[pos])) {
                        end = pos;
                        break;
                    }
                }

                pos++;
            }

            _currentPosition = pos;
            if (pos == _maxPosition) {
                end = _maxPosition;
            }

            if (start != end) {
                String token = new String(_charArray, start, end - start);
                if(wasEscaped) {
                    return _qqPattern.matcher(token).replaceAll(_qualifier);
                }
                return token;
            } else if (endQuotedString) {
                return EMPTY_STRING;
            } else {
                return null;
            }

        }

        // if delimiter more than 1 char long, make sure all chars match
        private boolean isDelimiter(int position) {
            boolean delimiterFound = true;
            for (int j = 0, J = _delimiters.length(); j < J; j++) {
                if (_delimiters.charAt(j) != _charArray[position++]) {
                    delimiterFound = false;
                    break;
                }
            }
            return delimiterFound;
        }

        // if qualifier more than 1 char long, make sure all chars match
        private boolean isQualifier(int position) {
            boolean qualifierFound = true;
            for (int j = 0, J = getQualifierLength(); j < J; j++) {
                if (getQualifier().charAt(j) != _charArray[position++]) {
                    qualifierFound = false;
                    break;
                }
            }
            return qualifierFound;
        }
    }
    
	// Expose default size of _lineCharArray for associated unit test.
    static final int LINE_CHAR_ARRAY_SIZE = 80;
    
    private String _fieldSep;
    private String _preferredLineSep;
    private String[] _lineSeps;
    private char[] _lineCharArray = new char[LINE_CHAR_ARRAY_SIZE];
    private String _qualifier;
    private Pattern _qPattern;
    private Pattern _qqPattern;
}